// -*- C++ -*-
/*!
 * @file  Robot_position.cpp
 * @brief ModuleDescription
 * @date $Date$
 *
 * $Id$
 */

#include "Robot_position.h"
#define M_PI            3.14159265358979323846

// Module specification
// <rtc-template block="module_spec">
static const char* robot_position_spec[] =
  {
    "implementation_id", "Robot_position",
    "type_name",         "Robot_position",
    "description",       "ModuleDescription",
    "version",           "1.0.0",
    "vendor",            "Yamamoto",
    "category",          "Category",
    "activity_type",     "EVENTDRIVEN",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.LRFdata_min", "84",
    "conf.default.LRFdata_mid", "347",
    "conf.default.LRFdata_max", "610",
    "conf.default.right_corner", "271",
    "conf.default.left_corner", "423",
    "conf.default.LRF_total_steps", "1024",
    "conf.default.LRFdata_theta", "360",
    "conf.default.robot_width", "50",
    "conf.default.field_width", "1000",
    "conf.default.field_depth", "2000",
    "conf.default.robot_board_distance", "100",

    // Widget
    "conf.__widget__.LRFdata_min", "text",
    "conf.__widget__.LRFdata_mid", "text",
    "conf.__widget__.LRFdata_max", "text",
    "conf.__widget__.right_corner", "text",
    "conf.__widget__.left_corner", "text",
    "conf.__widget__.LRF_total_steps", "text",
    "conf.__widget__.LRFdata_theta", "text",
    "conf.__widget__.robot_width", "text",
    "conf.__widget__.field_width", "text",
    "conf.__widget__.field_depth", "text",
    "conf.__widget__.robot_board_distance", "text",
    // Constraints

    "conf.__type__.LRFdata_min", "int",
    "conf.__type__.LRFdata_mid", "int",
    "conf.__type__.LRFdata_max", "int",
    "conf.__type__.right_corner", "int",
    "conf.__type__.left_corner", "int",
    "conf.__type__.LRF_total_steps", "int",
    "conf.__type__.LRFdata_theta", "int",
    "conf.__type__.robot_width", "int",
    "conf.__type__.field_width", "int",
    "conf.__type__.field_depth", "int",
    "conf.__type__.robot_board_distance", "int",

    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
Robot_position::Robot_position(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_range_dataIn("range_data", m_range_data),
    m_position_XOut("position_X", m_position_X),
    m_position_YOut("position_Y", m_position_Y),
    m_directionOut("direction", m_direction)

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
Robot_position::~Robot_position()
{
}



RTC::ReturnCode_t Robot_position::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("range_data", m_range_dataIn);
  
  // Set OutPort buffer
  addOutPort("position_X", m_position_XOut);
  addOutPort("position_Y", m_position_YOut);
  addOutPort("direction", m_directionOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("LRFdata_min", m_LRFdata_min, "84");					//LRF̍ŏXebvԍ
  bindParameter("LRFdata_mid", m_LRFdata_mid, "347");					//LRF̒ԃXebvԍ
  bindParameter("LRFdata_max", m_LRFdata_max, "610");					//LRF̍őXebvԍ
  bindParameter("right_corner", m_right_corner, "271");					//͈(tB[h)̉Ep̃Xebvԍ
  bindParameter("left_corner", m_left_corner, "423");					//tB[h̍p̃Xebvԍ
  bindParameter("LRF_total_steps", m_LRF_total_steps, "1024");			//LRF̃g[^Xebv
  bindParameter("LRFdata_theta", m_LRFdata_theta, "360");				//pxipx\ZoɎgĂpxj
  bindParameter("robot_width", m_robot_width, "50");
  bindParameter("field_width", m_field_width, "1000");					//tB[h̕(x >= 0)
  bindParameter("field_depth", m_field_depth, "2000");					//tB[h̉s
  bindParameter("robot_board_distance", m_robot_board_distance, "100");	//{bg{̂ƃ{bg㕔̔̋
  // </rtc-template>
  
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t Robot_position::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Robot_position::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Robot_position::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t Robot_position::onActivated(RTC::UniqueId ec_id)
{
	for (int i = m_LRFdata_min; i < m_LRFdata_max; i++) {
		theta = ((i - m_LRFdata_mid) * m_LRFdata_theta / 360 * 2 * M_PI / m_LRF_total_steps)+(M_PI / 2);   //WAPʂ̊pxZo

		if (i >= m_LRFdata_min && i < m_right_corner)field[i] =m_field_width / cos(theta);		//`͈̑(tB[h)̌vZ
		else if (i >= m_right_corner&&i < m_left_corner)field[i] =m_field_depth / sin(theta);
		else if (i >= m_left_corner &&i < m_LRFdata_max)field[i] = -m_field_width / cos(theta);
		//printf("theta:%f i:%d  l:%f\n", theta*180/M_PI, i, field[i]);
	}
	calc_count = 0;
  return RTC::RTC_OK;
}


RTC::ReturnCode_t Robot_position::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}


RTC::ReturnCode_t Robot_position::onExecute(RTC::UniqueId ec_id)
{
	if (m_range_dataIn.isNew()) {
		m_range_dataIn.read();	//URGf[^ǂݍ
		for (int i = m_LRFdata_min; i < m_LRFdata_max; i++) {
			range[i] = m_range_data.ranges[i];	//ϐɃf[^
			if (range[i] < 20)range[i] = 4000;	//URG͈̑͊O邢͑G[
			//printf("i:%d range:%f\n", i, range[i]);
		}

		right_count = left_count = 0;
		for (int i = m_LRFdata_min + 1; i < m_LRFdata_max - 1; i++) {

			if (range[i] < field[i]) {			//tB[hɂ郌[U

				a = range[i - 1] - range[i];	//Ẽׂ[UƂ̋
				b = range[i] - range[i + 1];	//ׂ̃[UƂ̋

				if (a > 100 && fabs(b) < 40) {	//Ώۃ[ỦEׂׂ̋̋Ɣב傫Ƃ,E[Ƃ݂Ȃ

					right_count += 1;
					right_theta = ((i - m_LRFdata_mid) * m_LRFdata_theta / 360 * 2 * M_PI / m_LRF_total_steps) + (M_PI / 2);	//WAPʂ̊pxZo
					right_x = range[i] * cos(right_theta);	//ɍW烏[hWւ̕ϊ
					right_y = range[i] * sin(right_theta);
					//printf("right i:%d range:%f theta:%f x:%f y:%f count:%d\n", i, range[i], right_theta * 180 / M_PI, right_x, right_y, right_count);
				}
				else if (b < -100 && fabs(a) < 40) {	//[

					left_count += 1;
					left_theta = ((i - m_LRFdata_mid) * m_LRFdata_theta / 360 * 2 * M_PI / m_LRF_total_steps) + (M_PI / 2);
					left_x = range[i] * cos(left_theta);
					left_y = range[i] * sin(left_theta);
					//printf("left  i:%d range:%f theta:%f x:%f y:%f count:%d\n", i, range[i], left_theta * 180 / M_PI, left_x, left_y, left_count);
				}
			}
		}
		if (right_count == 1 && left_count == 1) {	//E[ƍ[𔭌
			middle_x = (right_x + left_x) / 2;	//_̍W
			middle_y = (right_y + left_y) / 2;

			robot_direction = (atan((right_y - left_y) / (right_x - left_x))) + M_PI / 2;	//{bgĂpx

			position_x = middle_x + m_robot_board_distance*cos(robot_direction);		//{bg̍W
			position_y = middle_y + m_robot_board_distance*sin(robot_direction);
			printf("theta:%f, x:%f, y:%f\n", robot_direction * 180 / M_PI, position_x, position_y);

			//P̃f[^𑗂ꍇ
			/*
			m_position_X.data = position_x;			//{bgR|[lg֏
			m_position_Y.data = position_y;
			m_direction.data = robot_direction;

			m_position_XOut.write();
			m_position_YOut.write();
			m_directionOut.write();
		}*/

			//̃f[^𕽋ςꍇ
			
			position_array_x[calc_count] = position_x;	//zɃf[^𕡐W߂
			position_array_y[calc_count] = position_y;
			direction_array[calc_count] = robot_direction;

		}
		//else if (right_count == 0 && left_count == 0) printf("not found\n");	
		else {
			if (right_count == 0 && left_count == 0) printf("not found\n");	//tB[hɕ̂Ȃꍇ
			else printf("Detect multiple objects\n");						//tB[hɕ̂ꍇ
		
			position_array_x[calc_count] = -1;
			position_array_y[calc_count] = -1;
			direction_array[calc_count] = -1;
		}
		calc_count++;

		if (calc_count == samples_number)			//萔f[^܂ۂɕς
		{
			calc_count = 0;

			position_x = Average(position_array_x, samples_number);
			position_y = Average(position_array_y, samples_number);
			robot_direction = Average(direction_array, samples_number);

			if (position_x != 0 && position_y != 0) {	//f[^萔܂Ăꍇ̂ݏ݂s
				m_position_X.data = position_x;			//ϐOutPort֑
				m_position_Y.data = position_y;
				m_direction.data = robot_direction;

				m_position_XOut.write();				//{bgR|[lg֏
				m_position_YOut.write();
				m_directionOut.write();
				printf("\nOutport: theta:%f, x:%f, y:%f\n\n", robot_direction * 180 / M_PI, position_x, position_y);
			}
		}
	}
	return RTC::RTC_OK;
}

	/*
	RTC::ReturnCode_t Robot_position::onAborting(RTC::UniqueId ec_id)
	{
	  return RTC::RTC_OK;
	}
	*/

	/*
	RTC::ReturnCode_t Robot_position::onError(RTC::UniqueId ec_id)
	{
	  return RTC::RTC_OK;
	}
	*/

	/*
	RTC::ReturnCode_t Robot_position::onReset(RTC::UniqueId ec_id)
	{
	  return RTC::RTC_OK;
	}
	*/

	/*
	RTC::ReturnCode_t Robot_position::onStateUpdate(RTC::UniqueId ec_id)
	{
	  return RTC::RTC_OK;
	}
	*/

	/*
	RTC::ReturnCode_t Robot_position::onRateChanged(RTC::UniqueId ec_id)
	{
	  return RTC::RTC_OK;
	}
	*/

double Robot_position::Average(const double arr[], int length)
{
	int j = 0;
	for (int i = 0; i < length; i++) {
		if (arr[i] != -1)j++;
	}
	double sum = 0;
	if (j >= length - 2) {						//f[^萔܂ꍇ̂ݕς߂
		for (int i = 0; i < length; i++) {
			if (arr[i] != -1)sum += arr[i];
		}
		return sum / j;							//
	}
	else return 0;								//f[^萔Ȃꍇ0Ԃ
}

extern "C"
{
 
  void Robot_positionInit(RTC::Manager* manager)
  {
    coil::Properties profile(robot_position_spec);
    manager->registerFactory(profile,
                             RTC::Create<Robot_position>,
                             RTC::Delete<Robot_position>);
  }
  
};


